# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from .display import ChartDisplay
from .mpld3ChartDisplay import Mpld3ChartDisplay
import matplotlib.pyplot as plt
import numpy as np
import mpld3
import pandas as pd
from itertools import product
from .plugins.dialog import DialogPlugin
from pyspark import SparkContext
from pyspark.sql import functions as F
from pyspark.sql import SQLContext
from random import randint

class BarChartDisplay(Mpld3ChartDisplay):
    
    def getChartContext(self, handlerId):
        return ('barChartOptionsDialogBody.html', {})

    def getKeyFieldLabels(self, handlerId, aggregation, keyFields):
        stacked = self.options.get("stacked", "true") == "true"
        grouped = not stacked
        if len(keyFields)>1 and grouped:
            df = self.entity.groupBy(keyFields[0]).count().sort(F.col(keyFields[0]).asc()).dropna()        
            maxRows = int(self.options.get("rowCount","100"))
            numRows = min(maxRows,df.count())
            rows = df.take(numRows)
            labels = []
            for i, row in enumerate(rows):
                labels.append(str(row[keyFields[0]]))
            return labels
        return super(BarChartDisplay,self).getKeyFieldLabels(handlerId,aggregation,keyFields)
    
    def doRenderMpld3(self, handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues):
        stacked = self.options.get("stacked", "true") == "true"
        grouped = not stacked
        if len(keyFields)>1 and grouped:
            self.generateGroupedSeries(handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues)
        elif grouped == False and len(valueFields)>1:
            self.generateStackedBarChart(handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues);            
        else:
            self.generateBarChart(handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues);

    def generateBarChart(self, handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues):
        numColumns = len(keyFieldValues)
        barWidth = min(0.35, 0.9/len(valueFields))
        x_intv = np.arange(numColumns)
        for i, valueField in enumerate(valueFields):
            bar = ax.bar(x_intv+(i*barWidth), valueFieldValues[i], barWidth, color=colormap(1.*i/numColumns), alpha=0.5, label=valueField)
            self.connectElementInfo(bar, valueFieldValues[i])
        plt.xticks(x_intv+(barWidth/2),keyFieldLabels)
        plt.xlabel(", ".join(keyFields), fontsize=18)

    def generateStackedBarChart(self, handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues):
        numColumns = len(keyFieldValues)
        barWidth = 0.35
        x_intv = np.arange(numColumns)
        bar = ax.bar(x_intv, valueFieldValues[0], barWidth, alpha=0.5, label=valueFields[0])
        self.connectElementInfo(bar, valueFieldValues[0])
        colors= ['#79c36a','#f1595f','#599ad3','#f9a65a','#9e66ab','#cd7058','#d77fb3','#727272']
        bottom=valueFieldValues[0]
        for i in range(1,len(valueFields)):
            bar = plt.bar(x_intv,valueFieldValues[i],barWidth,label=valueFields[i],color=colors[i],bottom=bottom)          
            self.connectElementInfo(bar, valueFieldValues[i])
        bottom = self.sumzip(bottom,valueFieldValues[i])

    def generateGroupedSeries(self, handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues):
        def safeRepr(o):
            import decimal
            if isinstance(o, decimal.Decimal):
                return float(o)
            return o
        # convert to pandas
        pdf = self.entity.toPandas()
        # fill in gaps
        vals = []
        for keyField in keyFields:
            vals.append(pdf[keyField].unique())
        pdf2 = pd.DataFrame(list(product(*vals)), columns=keyFields)
        pdf = pd.merge(pdf,pdf2, on=keyFields, how='outer').fillna(0)
        # convert back to sql dataframe (fillna required)
        df = SQLContext(SparkContext.getOrCreate()).createDataFrame(pdf).fillna(0)
        series = df.rdd.map(lambda r:(safeRepr(r[0]), [( safeRepr(r[1]), safeRepr(r[valueFields[0]]))]))\
            .reduceByKey(lambda x,y: x+y)\
            .map(lambda r: (r[0], sorted(r[1], key=lambda tup: tup[0])))\
            .sortByKey()\
            .collect()
        maxLen=reduce(lambda x,y: max(x, len(y[1])), series, 0)
        ind=np.arange(len(series))
        #FIXME: generate random colors
        colors= ['#79c36a','#f1595f','#599ad3','#f9a65a','#9e66ab','#cd7058','#d77fb3','#727272']
        for i in range(maxLen):
            data=[t[1][i][1] if i<len(t[1]) else 0 for t in series]
            bars = ax.bar(ind + (i*0.15), data, width=0.15, color=colors[i], label=series[0][1][i][0])
            self.connectElementInfo(bars, data)
        plt.xticks(ind+0.3, [str(x[0]) for x in series])
        plt.ylabel(valueFields[0])
        plt.xlabel(keyFields[0])

        self.titleLegend=keyFields[1]

    def sumzip(self,x,y):
        return [sum(values) for values in zip(x,y)]



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
import os
import subprocess
import re
from IPython.core.magic import (Magics, magics_class, cell_magic)
from pixiedust.utils.javaBridge import *
from pixiedust.utils.template import *

'''
Manages the variables defined interactively in the Notebook
'''
class InteractiveVariables(object):
    def __init__(self, shell):
        self.shell = shell

    def getVar(self, varName):
        return self.shell.user_ns.get(varName, None ) or self.shell.user_ns_hidden.get(varName, None)

    def varTypeTransformer(self, varName, varValue):
        pythonToScalaSimpleTypeMap = {"str":"String","int":"Int"}
        scalaType = pythonToScalaSimpleTypeMap.get(varValue.__class__.__name__, None)
        if scalaType == "String":
            varValue = "\"" + varValue.replace('\n','\\n') + "\""
        return {"value": varValue, "codeValue": varValue if scalaType is not None else None, "type": scalaType or "Any"}

    def getVarsDict(self):
        user_ns = self.shell.user_ns
        user_ns_hidden = self.shell.user_ns_hidden
        nonmatching = object()  # This can never be in user_ns
        #for i in user_ns:
        #    print(user_ns[i].__class__)
        filtered = ["function"]
        out = { key : self.varTypeTransformer(key, user_ns[key]) for key in user_ns \
                if not key.startswith('_') \
                and key!="sc" and key!="sqlContext" \
                and (user_ns[key] is not user_ns_hidden.get(key, nonmatching)) \
                and not inspect.isclass(user_ns[key])\
                and not inspect.isfunction(user_ns[key])\
                and not inspect.ismodule(user_ns[key])}
        return out

    def updateVarsDict(self, vars):
        self.shell.user_ns.update(vars)

@magics_class
class PixiedustScalaMagics(Magics):
    def __init__(self, shell):
        super(PixiedustScalaMagics,self).__init__(shell)
        self.interactiveVariables = InteractiveVariables(shell)
        self.scala_home = os.environ.get("SCALA_HOME")
        self.class_path = JavaWrapper("java.lang.System").getProperty("java.class.path")
        self.env = PixiedustTemplateEnvironment()

    def getLineOption(self, line, optionName):
        m=re.search(r"\b" + optionName + r"=(\S+)",line)
        return m.group(1) if m is not None else None

    def hasLineOption(self, line, optionName):
        return re.match(r"\b" + optionName + r"\b", line) is not None

    def getReturnVars(self, code):
        vars=set()
        for m in re.finditer(r"\b__(\w+?)\b",code):
            vars.add(m.group(0))
        return vars

    def fromJava(self, stuff):
        if stuff.__class__.__name__ == "JavaObject":
            if stuff.getClass().getName() == "org.apache.spark.sql.DataFrame":
                return DataFrame(stuff, SQLContext(SparkContext.getOrCreate(), stuff.sqlContext()))
            elif stuff.getClass().getName() == "org.apache.spark.sql.SQLContext":
                return SQLContext(SparkContext.getOrCreate(),stuff)
        return stuff

    @cell_magic
    def scala(self, line, cell):
        if not self.scala_home:
            print("Error Cannot run scala code: SCALA_HOME environment variable not set")
            return
        
        #generate the code
        scalaCode = self.env.getTemplate("scalaCell.template").render(
            cell=cell, variables=self.interactiveVariables.getVarsDict(), returnVars=self.getReturnVars(cell)
        )
        if self.hasLineOption(line, "debug"):
            print(scalaCode)
            return
        
        #build the scala object
        dir=os.path.expanduser('~') + "/pixiedust"
        if not os.path.exists(dir):
            os.makedirs(dir)
        source="pixiedustRunner.scala"
        with open(dir + "/" + source, "w") as f:
            f.write(scalaCode)
        #Compile the code
        proc = subprocess.Popen([self.scala_home + "/bin/scalac","-classpath", self.class_path, source],stdout=subprocess.PIPE,stderr=subprocess.PIPE, cwd=dir)
        code = proc.wait()
        if code != 0:
            while True:
                line = proc.stderr.readline()
                if not line:
                    break
                print(line.rstrip())
            return
        
        #Load the class and initialize the variables
        f = sc._jvm.java.io.File(dir)
        url = f.toURL()
        urls=sc._gateway.new_array(sc._jvm.java.net.URL,1)
        urls[0]=url

        cl = sc._jvm.java.net.URLClassLoader(urls)
        cls = sc._jvm.java.lang.Class.forName("com.ibm.pixiedust.PixiedustScalaRun$", True, cl)

        runnerObject = JavaWrapper(cls.getField("MODULE$").get(None), True, 
            self.getLineOption(line, "channel"), self.getLineOption(line, "receiver"))
        runnerObject.callMethod("init", pd_getJavaSparkContext(), self.interactiveVariables.getVar("sqlContext")._ssql_ctx )
        varMap = runnerObject.callMethod("runCell")

        #capture the return vars and update the interactive shell
        returnVars = {}
        it = varMap.iterator()
        while it.hasNext():
            t = it.next()
            returnVars[t._1()] = self.fromJava(t._2())
        self.interactiveVariables.updateVarsDict(returnVars)

        #discard the ClassLoader, we only use it within the context of a cell.
        #TODO: change that when we support inline scala class/object definition
        cl.close()
        cl=None
        cls=None
        runnerObject=None

try:
    get_ipython().register_magics(PixiedustScalaMagics)
except NameError:
    #IPython not available we must be in a spark executor
    pass



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from .mpld3ChartDisplay import Mpld3ChartDisplay
import math
import numpy as np

class PieChartDisplay(Mpld3ChartDisplay):

    def supportsKeyFieldLabels(self, handlerId):
        return False

    def supportsLegend(self, handlerId):
        return False
    
    def getPreferredDefaultValueFieldCount(self, handlerId):
		return 1

    def getDefaultAggregation(self, handlerId):
        return "COUNT"

    # override the default keys displayed when the chart first renders
    def getDefaultKeyFields(self, handlerId, aggregation):
        return self.sampleColumn((aggregation != "COUNT"))

    # override the default values displayed when the chart first renders
    def getDefaultValueFields(self, handlerId, aggregation):
        return self.getDefaultKeyFields(handlerId, aggregation)

    def setChartGrid(self, handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues):
        pass

    def doRenderMpld3(self, handlerId, fig, ax, colormap, keyFields, keyFieldValues, keyFieldLabels, valueFields, valueFieldValues):
        numPieCharts = len(valueFields)
        colors = colormap(np.linspace(0., 1., len(keyFieldValues)))
        if numPieCharts > 1:
            fig.delaxes(ax)
            if numPieCharts < 3:
                pieChartGridPrefix = "1" + str(numPieCharts)
            elif numPieCharts <= 4:
                pieChartGridPrefix = "22"
            else:
                pieChartGridPrefix = str(int(math.ceil(numPieCharts//3))) + "3"
            for i, valueField in enumerate(valueFields):
                ax2 = fig.add_subplot(pieChartGridPrefix + str(i+1))
                patches, texts, autotexts = ax2.pie(valueFieldValues[i], labels=keyFieldLabels, colors=colors, explode=None, autopct='%1.1f%%')
                for j, patch in enumerate(patches):
                    self.connectElementInfo(patch, valueFieldValues[i][j])
                ax2.set_title(valueFields[i], fontsize=18);
                ax2.axis("equal")
                # TODO: hide the x and y axis - this is not working
                ax2.get_xaxis().set_alpha(0)
                ax2.get_yaxis().set_alpha(0)
        else:
            patches, texts, autotexts = ax.pie(valueFieldValues[0], labels=keyFieldLabels, colors=colors, explode=None, autopct='%1.1f%%')
            for j, patch in enumerate(patches):
                self.connectElementInfo(patch, valueFieldValues[0][j])
            ax.set_title(valueFields[0], fontsize=18);
            ax.axis("equal")
            # TODO: hide the x and y axis - this is not working
            ax.get_xaxis().set_alpha(0)
            ax.get_yaxis().set_alpha(0)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from __future__ import print_function
import sqlite3
import sys
import platform
import urllib
import os
from pixiedust.utils.storage import *
from pixiedust.utils.printEx import *
from maven import Artifact
from maven import downloader
from maven import RequestException
from pyspark import SparkContext

old_gen_filename=Artifact._generate_filename
def myGenFileName(self):
    if self.version:
        s = self.artifact_id + "-" + self.version
        if not self.classifier:
            return  s + "." + self.extension
        else:
            return s + "-" + self.classifier + "." + self.extension 
    else:
        return old_gen_filename(self)
            
Artifact._generate_filename=myGenFileName

PACKAGES_TBL_NAME="SPARK_PACKAGES"

class __ArtifactStorage(Storage):
    def __init__(self):
        self._initTable( PACKAGES_TBL_NAME,
        '''
            GROUPID        TEXT  NOT NULL,
            ARTIFACTID     TEXT  NOT NULL,
            VERSION        TEXT  NOT NULL,
            FILEPATH       TEXT  NOT NULL,
            BASE           TEXT,
            PRIMARY KEY (GROUPID, ARTIFACTID)
        ''')

artifactStorage = __ArtifactStorage()

class PackageManager(object):
    DOWNLOAD_DIR=os.path.expanduser('~') + "/data/libs"
    
    def __init__(self):
        if not os.path.exists(self.DOWNLOAD_DIR):
            os.makedirs(self.DOWNLOAD_DIR)

    def _toArtifact(self, artifact):
        if isinstance(artifact, basestring):
            #check if the user wants a direct download
            if artifact.startswith("http://") or artifact.startswith("https://") or artifact.startswith("file://"):
                url=artifact
                artifact=type("",(),
                    {
                        "group_id":"direct.download",
                        "artifact_id": artifact,
                        "version":"1.0",
                        "get_filename":lambda self,dir: os.path.join(dir, url.split("/")[-1] ),
                        "is_snapshot": lambda self:False,
                        "with_version": lambda self,version:self,
                        "uri": lambda self, base, version: url,
                        "__str__": lambda self: url
                    }
                )()
            else:
                artifact=Artifact.parse(artifact)
        return artifact
    
    def installPackage(self, artifact, base=None, sc=None):
        artifact = self._toArtifact(artifact)
        #Test if we already have a version installed
        res=self.fetchArtifact(artifact)
        fileLoc=None
        if res:
            fileLoc=res[1]
            print("Package already installed: {0}".format(str(artifact)))
        else:
            #download package
            art=[artifact]
            def _doDownload(d):
                artifact=art[0]
                if not artifact.version or artifact.version=='0':
                    artifact.version = d.resolver._find_latest_version_available(artifact)
                fileLoc = artifact.get_filename(self.DOWNLOAD_DIR)
                if os.path.isfile(fileLoc):
                    os.remove(fileLoc)
                results = d.download(artifact,filename=self.DOWNLOAD_DIR)
                if not results[1]:
                    raise Exception("Error downloading package {0}".format(str(artifact)))
                else:
                    artifact=results[0]
                    print("Artifact downloaded successfully {0}".format(str(artifact)))
                    printEx("Please restart Kernel to complete installation of the new package",PrintColors.RED)
                fileLoc=self.storeArtifact(artifact,base)
                return fileLoc
            
            try:
                fileLoc=_doDownload(downloader.Downloader(base) if base is not None else downloader.Downloader())
            except RequestException as e:
                #try another base
                try:
                    fileLoc=_doDownload(downloader.Downloader("http://dl.bintray.com/spark-packages/maven"))
                except RequestException as e:
                    print("Unable to install artifact {0}".format(e.msg))
                    raise
            except:
                print(str(sys.exc_info()[1]))
                raise
        if sc is None:
            sc = SparkContext.getOrCreate()
            
        if sc:
            #convert to file uri for windows platform
            if platform.system()=='Windows':
                fileLoc="file://" + urllib.pathname2url(fileLoc)
            sc.addPyFile(fileLoc)
            
        return artifact
        
    def uninstallPackage(self, artifact):
        results = self.fetchArtifact(artifact)
        if results is None:
            print("Package is not installed: {0}".format(str(artifact)))
        else:
            self._deleteArtifact(results[0],results[1])
    
    def hasArtifact(self, artifact):
        return self.fetchArtifact(artifact) is not None
        
    def fetchArtifact(self, artifact ):
        artifact = self._toArtifact(artifact)
        return artifactStorage.fetchOne("""
                SELECT * from {0} WHERE GROUPID='{1}' and ARTIFACTID='{2}'
            """.format(PACKAGES_TBL_NAME,artifact.group_id,artifact.artifact_id), 
            lambda row: (Artifact(row["GROUPID"],row["ARTIFACTID"],row["VERSION"]),row["FILEPATH"])
        )
                
    def printAllPackages(self):
        self.visitAll(lambda row: print("{0}:{1}:{2} => {3}".format(row["GROUPID"],row["ARTIFACTID"],row["VERSION"],row["FILEPATH"])))
        
    def visitAll(self, walker):
        artifactStorage.execute("""
                SELECT * FROM {0}
            """.format(PACKAGES_TBL_NAME),
            walker
        )
        
    def storeArtifact(self, artifact,base=None):
        artifact = self._toArtifact(artifact)
        fileLoc=artifact.get_filename(self.DOWNLOAD_DIR)
        artifactStorage.insert("""
            INSERT INTO {0} (GROUPID,ARTIFACTID,VERSION,BASE,FILEPATH)
            VALUES ('{1}','{2}','{3}','{4}','{5}')
        """.format(
                PACKAGES_TBL_NAME,
                artifact.group_id,
                artifact.artifact_id,
                artifact.version,
                base if base is not None else "",
                fileLoc
            )
        )
        print("Successfully added artifact {0}".format(str(artifact)))
        return fileLoc
        
    def _deleteArtifact(self,artifact,filePath):
        artifact = self._toArtifact(artifact)
        rowDeleted = artifactStorage.delete("""
            DELETE FROM {0} WHERE GROUPID='{1}' AND ARTIFACTID='{2}'
        """.format(
                PACKAGES_TBL_NAME,
                artifact.group_id,
                artifact.artifact_id
            )
        )
        if rowDeleted==0:
            print("Artifact {0} not deleted because it doesn't exist".format(str(artifact)))
        else:
            print("Successfully deleted artifact {0}".format(str(artifact)))
            
        if filePath is not None:
            os.remove(filePath)
    